/* threatmateState.h
 */
#ifndef SEARCH_THREATMATESTATE_H
#define SEARCH_THREATMATESTATE_H

#include "osl/misc/carray.h"
#include "osl/move.h"
#ifdef OSL_SMP
#  include "osl/misc/lightMutex.h"
#endif
#include <iosfwd>
namespace osl
{
  namespace search
  {
    struct ThreatmateStateData
    {
      Move threatmate_move;
      char current_status;
    };
    /**
     * ͤθ;õΤξִ.
     * http://www31.ocn.ne.jp/~kfend/inside_kfend/ptc.html#c3
     *
     * - (MAYBE_)THREATMATE --(check)--> CHECK_AFTER_THREATMATE 
     * - (MAYBE_)THREATMATE --(no-check)--> MAY_HAVE_CHECKMATE
     * - CHECK_AFTER_THREATMATE --(escape)-->  MAYBE_THREATMATE
     */
    class ThreatmateState : protected ThreatmateStateData
    {
    public:
      enum Status {
	UNKNOWN = 0,
	/** threatmate found by checkmate search */
	THREATMATE,
	/** threatmate, not sure */
	MAYBE_THREATMATE,
	/** status after threatmate responded by check */
	CHECK_AFTER_THREATMATE,
	/** status after threatmate responded by non-check move */
	MAY_HAVE_CHECKMATE,
      };
    private:
      static const CArray<Status,5*2> transition;
#ifdef OSL_SMP
      mutable LightMutex mutex;
#endif
    public:
      ThreatmateState(Status s=UNKNOWN)
      {
	current_status = s;
      }
      ThreatmateState(const ThreatmateState& src) : ThreatmateStateData(src)
      {
      }
      ThreatmateState& operator=(const ThreatmateState& src)
      {
	if (this != &src)
	  ThreatmateStateData::operator=(src);
	return *this;
      }
      void setThreatmate(Status s) { 
#ifdef OSL_USE_RACE_DETECTOR
	LightMutex::scoped_lock lk(mutex);
#endif
	current_status = s; 
      }
      void setThreatmate(Move m) { 
#ifdef OSL_SMP
	LightMutex::scoped_lock lk(mutex);
#endif
	threatmate_move = m;
	current_status = THREATMATE; 
      }

      bool isUnknown() const { 
#ifdef OSL_USE_RACE_DETECTOR
	LightMutex::scoped_lock lk(mutex);
#endif
	return current_status == UNKNOWN; 
      }
      bool isThreatmate() const { 
#ifdef OSL_USE_RACE_DETECTOR
	LightMutex::scoped_lock lk(mutex);
#endif
	return current_status == THREATMATE; 
      }
      bool maybeThreatmate() const { 
#ifdef OSL_USE_RACE_DETECTOR
	LightMutex::scoped_lock lk(mutex);
#endif
	return (current_status == THREATMATE) 
	  || (current_status == MAYBE_THREATMATE);
      }
      bool mayHaveCheckmate() const {
#ifdef OSL_USE_RACE_DETECTOR
	LightMutex::scoped_lock lk(mutex);
#endif
	return current_status == MAY_HAVE_CHECKMATE; 
      }
      Status status() const { 
#ifdef OSL_USE_RACE_DETECTOR
	LightMutex::scoped_lock lk(mutex);
#endif
	return static_cast<Status>(current_status); 
      }
      const Move threatmateMove() const { 
#ifdef OSL_USE_RACE_DETECTOR
	LightMutex::scoped_lock lk(mutex);
#endif
	return threatmate_move; 
      }
      const ThreatmateState newStatus(bool is_check) const
      {
#ifdef OSL_USE_RACE_DETECTOR
	LightMutex::scoped_lock lk(mutex);
#endif
	return transition[current_status*2+is_check];
      }

      void update(const ThreatmateState *parent, bool in_check)
      {
	if (maybeThreatmate() || ! parent)
	  return;
	const ThreatmateState new_status = parent->newStatus(in_check);
#ifdef OSL_SMP
	LightMutex::scoped_lock lk(mutex);
	if (! (current_status == THREATMATE
	       || current_status == MAYBE_THREATMATE))
#endif
	  *this = new_status;
      }
    };
    std::ostream& operator<<(std::ostream&, ThreatmateState);
  } // namespace search
} // osl

#endif /* SEARCH_THREATMATESTATE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
