/* effectUtil.cc
 */
#include "osl/effect_util/effectUtil.h"
#include "osl/effect_util/effectUtil.tcc"
#include "osl/eval/pieceEval.h"
#include "osl/eval/progressEval.h"
#include "osl/eval/ml/testEval.h"
#include "osl/state/numEffectState.h"

osl::Piece osl::effect_util::EffectUtil::
cheapestPiece(Player P, const NumEffectState& state, Position pos)
{
  assert(pos.isOnBoard());
  effect_action::SelectCheapest selector;
  forEachEffect(P, state, pos, selector);
  return selector.result;
}

#ifndef MINIMAL
osl::Piece osl::effect_util::EffectUtil::
promotablePiece(Player P, const NumEffectState& state, 
		Position effectTarget)
{
  assert(effectTarget.isOnBoard());
  effect_action::SelectPromotable selector;
  forEachEffect(P, state, effectTarget, selector);
  return selector.result;
}
void osl::effect_util::EffectUtil::showEffect(const NumEffectState& state, Position target,
				 std::ostream& os)
{
  assert(target.isOnBoard());
  PieceVector pieces;
  findEffect(BLACK, state, target, pieces);
  os << pieces;
  pieces.clear();
  findEffect(WHITE, state, target, pieces);
  os << pieces;
}
#endif
void
osl::effect_util::EffectUtil::
findEffect(Player P, const NumEffectState& state, Position target,
	   PieceVector& out)
{
  effect_action::StorePiece store(&out);
  forEachEffect(P, state, target, store);
}

namespace osl
{
#ifndef MINIMAL
  template bool
  EffectUtil::isOpenAttackMove<BLACK>(const NumEffectState&, Position, Position, Position);
  template bool
  EffectUtil::isOpenAttackMove<WHITE>(const NumEffectState&, Position, Position, Position);
  template bool
  EffectUtil::isAddEffectMove<BLACK>(const NumEffectState&, Position, Move);
  template bool
  EffectUtil::isAddEffectMove<WHITE>(const NumEffectState&, Position, Move);

  template void
  EffectUtil::findThreat<PieceEval>(const NumEffectState& state,
				    Position position,
				    PtypeO ptypeo,
				    PieceVector& out);
  template void
  EffectUtil::findThreat<osl::eval::ProgressEval>(const NumEffectState& state,
						  Position position,
						  PtypeO ptypeo,
						  PieceVector& out);
#endif
  template void
  EffectUtil::findThreat<osl::eval::ml::TestEval>(
    const NumEffectState& state,
    Position position,
    PtypeO ptypeo,
    PieceVector& out);
  template Piece 
  EffectUtil::safeCaptureNotByKing<BLACK>(NumEffectState const&, Position, Piece);
  template Piece 
  EffectUtil::safeCaptureNotByKing<WHITE>(NumEffectState const&, Position, Piece);
}

#ifndef MINIMAL
struct osl::effect_util::EffectUtil::SelectMostValuable
{
  const NumEffectState& state;
  Player target;
  Position move_from;
  Piece result;
  int result_value;
  SelectMostValuable(const NumEffectState& s, Player t, Position f) 
    : state(s), target(t), move_from(f),
      result(Piece::EMPTY()), result_value(0)
  {
  }
  void operator()(Position pos)
  {
    if (pos == move_from)
      return;
    const Piece cur = state.getPieceOnBoard(pos);
    assert(cur.isPiece());
    if (cur.owner() == target)
    {
      const Ptype ptype = cur.ptype();
      const int curValue 
	= eval::Ptype_Eval_Table.captureValue(newPtypeO(WHITE,ptype));
      assert(curValue > 0);
      if ((ptype != KING) && (curValue > result_value))
      {
	result = cur;
	result_value = curValue;
      }
    }
  }
};

osl::Piece osl::EffectUtil::
mostValuableEffectOfMoveNotKing(Player target, const NumEffectState& state, Move m)
{
  SelectMostValuable selector(state, target, m.from());
  forEachEffectPieceOfMove(state, m, selector);
  return selector.result;
}

template <int N>
struct osl::effect_util::EffectUtil::SelectMostValuableN
{
  const NumEffectState& state;
  Player target;
  Position move_from;
  CArray<Piece,N>& result;
  CArray<int,N> result_value;
  SelectMostValuableN(const NumEffectState& s, Player t, Position f,
		      CArray<Piece,N>& out)
    : state(s), target(t), move_from(f), result(out)
  {
    result.fill(Piece::EMPTY());
    result_value.fill(0);
  }
  void operator()(Position pos)
  {
    if (pos == move_from)
      return;
    const Piece cur = state.getPieceOnBoard(pos);
    assert(cur.isPiece());
    if (cur.owner() == target)
    {
      const Ptype ptype = cur.ptype();
      const int curValue
	= eval::Ptype_Eval_Table.captureValue(newPtypeO(WHITE,ptype));
      assert(curValue > 0);
      if (ptype == KING)
	return;
      for (int i=0; i<N; ++i)
      {
	if (curValue > result_value[i])
	{
	  for (int j=i; j+1<N; ++j)
	  {
	    result[j+1] = result[j];
	    result_value[j+1] = result_value[j];
	  }
	  result[i] = cur;
	  result_value[i] = curValue;
	  return;
	}
      }
    }
  }
};

void osl::EffectUtil::
mostValuableEffectOfMoveNotKing(Player target, const NumEffectState& state, 
				Move m, CArray<Piece,2>& out)
{
  SelectMostValuableN<2> selector(state, target, m.from(), out);
  forEachEffectPieceOfMove(state, m, selector);
}
#endif

template <class EvalT>
struct osl::effect_util::EffectUtil::FindThreat
{
  const NumEffectState& state;
  Player target;
  int attacker_value;
  PieceVector& supported, & unsupported;
  FindThreat(const NumEffectState& st, Player t, int a,
	     PieceVector& s, PieceVector& u)
    : state(st), target(t), attacker_value(a), supported(s), unsupported(u)
  {
  }
  void operator()(Position pos)
  {
    const Piece cur = state.getPieceOnBoard(pos);
    assert(cur.isPiece());
    if (cur.owner() != target)
      return;
    if (state.hasEffectBy(target, pos))
    {
      if (abs(EvalT::captureValue(cur.ptypeO()))
	  > attacker_value)
	supported.push_back(cur);
    }
    else
    {
      unsupported.push_back(cur);
    }
  }
};

template <class EvalT>
void osl::EffectUtil::
findThreat(const NumEffectState& state, Position position,
	   PtypeO ptypeo, PieceVector& out)
{
  PieceVector supported, unsupported;
  const int attacker_value = abs(EvalT::captureValue(ptypeo));
  FindThreat<EvalT> f(state, alt(getOwner(ptypeo)), attacker_value, 
		      supported, unsupported);
  forEachEffectOfPtypeO<FindThreat<EvalT>, false>
    (state, position, ptypeo, f);

  unsupported.sortByPtype();
  supported.sortByPtype();
  PieceVector::iterator u=unsupported.begin(), s=supported.begin();

  if (u!=unsupported.end())
  {
    while ((s!=supported.end()) 
	   && ((abs(EvalT::captureValue(s->ptypeO()))
		- attacker_value)
	       > abs(EvalT::captureValue(u->ptypeO()))))
    {
      out.push_back(*s);
      ++s;
    }
  }
  out.push_back(u, unsupported.end());
  out.push_back(s, supported.end());
}

#ifndef MINIMAL
osl::Ptype osl::EffectUtil::
cheapestPtype(Player P, const NumEffectState& state, Position effect_target)
{
  PieceMask pieces = state.getEffect(effect_target);
  pieces &= state.getOnBoardMask(P);
  Ptype result = PTYPE_EMPTY;
  // sorted by capture value
  if (pieces.none())
    return result;
  if ((pieces.getMask(PtypeFuns<PAWN>::indexNum) & mask_t::makeDirect(PtypeFuns<PAWN>::indexMask)).any()) {
    Position d = Board_Table.nextPosition(P, effect_target, D);
    if (state.getPieceAt(d).ptype() == PAWN)
      return PAWN;
    return PPAWN;
  }
#ifndef CHEAPEST_PTYPE_STRICT
  mask_t m = pieces.getMask(PtypeFuns<LANCE>::indexNum) & mask_t::makeDirect(PtypeFuns<LANCE>::indexMask);
  if (m.any()) {
    int num=m.takeOneBit()+(PtypeFuns<LANCE>::indexNum<<5);
    return state.getPieceOf(num).ptype();
  }
  m = pieces.getMask(0);
#  if OSL_WORDSIZE == 64
  return state.getPieceOf(m.takeOneBit()).ptype();
#  else
  if (m.any())
    return state.getPieceOf(m.takeOneBit()).ptype();
  m = pieces.getMask(1);
  return state.getPieceOf(m.takeOneBit()+32).ptype();
#  endif
#else
  mask_t m = pieces.getMask(PtypeFuns<LANCE>::indexNum) & mask_t::makeDirect(PtypeFuns<LANCE>::indexMask);
  while (m.any()) {
    int num=m.takeOneBit()+(PtypeFuns<LANCE>::indexNum<<5);
    if (state.getPieceOf(num).ptype() == LANCE)
      return LANCE;
    result = PLANCE;
  }
  m = pieces.getMask(PtypeFuns<KNIGHT>::indexNum) & mask_t::makeDirect(PtypeFuns<KNIGHT>::indexMask);
  while (m.any()) {
    int num=m.takeOneBit()+(PtypeFuns<KNIGHT>::indexNum<<5);
    if (state.getPieceOf(num).ptype() == KNIGHT)
      return KNIGHT;
    if (result == PTYPE_EMPTY)
      result = PKNIGHT;
  }
  m = pieces.getMask(PtypeFuns<SILVER>::indexNum) & mask_t::makeDirect(PtypeFuns<SILVER>::indexMask);
  while (m.any()) {
    int num=m.takeOneBit()+(PtypeFuns<SILVER>::indexNum<<5);
    if (state.getPieceOf(num).ptype() == SILVER)
      return SILVER;
    if (result == PTYPE_EMPTY)
      result = PSILVER;
  }
  if (result != PTYPE_EMPTY)
    return result;
  if ((pieces.getMask(PtypeFuns<GOLD>::indexNum) & mask_t::makeDirect(PtypeFuns<GOLD>::indexMask)).any())
    return GOLD;
  
  m = pieces.getMask(PtypeFuns<BISHOP>::indexNum) & mask_t::makeDirect(PtypeFuns<BISHOP>::indexMask);
  while (m.any()) {
    int num=m.takeOneBit()+(PtypeFuns<BISHOP>::indexNum<<5);
    if (state.getPieceOf(num).ptype() == BISHOP)
      return BISHOP;
    result = PBISHOP;
  }
  m = pieces.getMask(PtypeFuns<ROOK>::indexNum) & mask_t::makeDirect(PtypeFuns<ROOK>::indexMask);
  while (m.any()) {
    int num=m.takeOneBit()+(PtypeFuns<ROOK>::indexNum<<5);
    if (state.getPieceOf(num).ptype() == ROOK)
      return ROOK;
    if (result == PTYPE_EMPTY)
      result = PROOK;
  }
  return result;
#endif
}
#endif
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
