#include "osl/checkmate/dfpn.h"
#include "osl/record/csaString.h"
#include "osl/record/csaRecord.h"
#include "osl/state/hashEffectState.h"
#include "osl/misc/perfmon.h"
#include "osl/misc/milliSeconds.h"

#include "osl/move_generator/legalMoves.h"
#include "osl/apply_move/applyMove.h"
#include "osl/checkmate/dfpnRecord.h"

#include <boost/scoped_ptr.hpp>
#include <string>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <unistd.h>

#include <bitset>

using namespace osl;
using namespace osl::checkmate;
using namespace osl::misc;

bool verbose=false;
unsigned long long total_cycles=0;
bool show_escape_filename = false;
bool force_attack = false;
int num_checkmate=0, num_nocheckmate=0, num_escape=0, num_unkown=0;
double total_nodes=0, total_tables=0;
int limit = 100000;
bool blocking_verify = true;
bool debug = false;

void search(const char *filename);
void usage(const char *program_name)
{
  std::cerr << "usage: " << program_name << " [-v] [-f] [-l limit] csa-files\n";
}
int main(int argc, char **argv)
{
  const char *program_name = argv[0];
  bool error_flag = false;
  extern char *optarg;
  extern int optind;

  char c;
  while ((c = getopt(argc, argv, "fl:Vvh")) != EOF)
  {
    switch(c)
    {
    case 'f':	force_attack = true;
      break;
    case 'l':	limit = atoi(optarg);
      break;
    case 'V':	blocking_verify = false;
      break;
    case 'v':	verbose = true;
      break;
    default:	error_flag = true;
    }
  }
  argc -= optind;
  argv += optind;

  if (error_flag || (argc < 1)) {
    usage(program_name);
    return 1;
  }

  try
  {
    for (int i=0; i<argc; ++i)
    {
      search(argv[i]);
      total_cycles = 0;
    }
    std::cerr << "check " << num_checkmate << " nocheckmate " << num_nocheckmate << " escape " << num_escape
	      << " unknown " << num_unkown << "\n";
    std::cerr << "total nodes " << total_nodes 
	      << " tables " << total_tables << "\n";
  }
  catch (std::exception& e)
  {
    std::cerr << e.what() << "\n";
    return 1;
  }
}

/**
 * CSA Υե뤫 stream 1()ɤ
 */
void readCsaState(std::istream& is, SimpleState& state)
{
  Record rec;
  record::csa::InputStream irs(is);
  irs.load(&rec);
  state = rec.getInitialState();
}

double real_seconds = 0.0;

void analyzeCheckmate(Dfpn& searcher, const NumEffectState& state, Move checkmate_move)
{
  NumEffectState new_state = state;
  std::cerr << state << " " << checkmate_move << "\n";
  ApplyMoveOfTurn::doMove(new_state, checkmate_move);
  HashKey key(new_state);
  const DfpnTable& table = searcher.currentTable();
  DfpnRecordBase record = table.probe(key, PieceStand(WHITE, new_state));
  std::cerr << record.proof_disproof << " " << std::bitset<64>(record.solved) << "\n";

  MoveVector moves;
  LegalMoves::generate(new_state, moves);
  for (size_t i=0; i<moves.size(); ++i) {
    NumEffectState tmp = new_state;
    ApplyMoveOfTurn::doMove(tmp, moves[i]);
    DfpnRecordBase record = table.probe(key.newHashWithMove(moves[i]), PieceStand(WHITE, tmp));
    std::cerr << moves[i] << " " << record.proof_disproof << " " << record.best_move << "\n";
  }

  {
    MoveVector moves;
    if (state.getTurn() == BLACK)
      Dfpn::generateEscape<BLACK>(new_state, false, Position(), moves);
    else
      Dfpn::generateEscape<WHITE>(new_state, false, Position(), moves);
    std::cerr << "Escape " << moves;
    moves.clear();
    if (state.getTurn() == BLACK)
      Dfpn::generateEscape<BLACK>(new_state, true, Position(), moves);
    else
      Dfpn::generateEscape<BLACK>(new_state, true, Position(), moves);
    std::cerr << "Escape full " << moves;
  }
}

void testWinOrLose(const char *filename,
		   Dfpn& searcher,
		   const SimpleState& sstate, int limit)
{
  const Player P = sstate.getTurn();
  HashEffectState state((NumEffectState(sstate)));
  const PathEncoding path(state.getTurn());
  const Position my_king = state.getKingPosition(P);
  if ((! force_attack) 
      && ! my_king.isPieceStand() && state.inCheck(P))
  {
    // ꤫鲦꤬äƤ
    MilliSeconds timer = MilliSeconds::now();
    misc::PerfMon clock;
    ProofDisproof pdp = searcher.hasEscapeMove(state, state.getHash(), path, limit, Move::PASS(alt(P)));
    total_cycles += clock.stop();
    real_seconds = timer.elapsedSeconds();

    if (verbose)
      std::cerr << pdp << "\n";
    if (pdp.isCheckmateSuccess()) {
      ++num_checkmate;
    }
    else {
      if (pdp.isCheckmateFail()
	  /*|| record->findLoop(path, table.getTwinTable())*/)
	++num_escape;
      else {
	assert(! pdp.isFinal());
	++num_unkown;
      }
    }
    return;
  }

  Move checkmate_move;
  MilliSeconds timer = MilliSeconds::now();
  PerfMon clock;
  const ProofDisproof pdp = searcher.
    hasCheckmateMove(state, state.getHash(), path, limit, checkmate_move);
  total_cycles += clock.stop();
  real_seconds = timer.elapsedSeconds();
  if (verbose)
    std::cerr << pdp << "\n";

  if (pdp.isCheckmateSuccess()) {
    ++num_checkmate;
    if (verbose)
      std::cerr << checkmate_move << "\n";
    if (debug)
      analyzeCheckmate(searcher, state, checkmate_move);
  }
  else {
    if (pdp.isFinal()
	/*|| record->findLoop(path, table.getTwinTable())*/)
      ++num_nocheckmate;
    else
      ++num_unkown;
  }
}

void search(const char *filename)
{
  std::ifstream is(filename);
  if (! is) {
    std::cerr << "\nskipping " << filename << "\n";
    return;
  }
  if (verbose)
    std::cerr << "\nsolving " << filename << "\n";
  
  Dfpn searcher;
  searcher.setBlockingVerify(blocking_verify);
  SimpleState state;
  readCsaState(is, state);
  NumEffectState nstate(state);
  const bool attack = force_attack
    || state.getKingPosition(state.getTurn()).isPieceStand()
    || ! nstate.inCheck(state.getTurn());
  DfpnTable table(attack ? state.getTurn() : alt(state.getTurn()));
  searcher.setTable(&table);
  testWinOrLose(filename, searcher, state, limit);
  const size_t table_used = searcher.currentTable().size();
  total_nodes += searcher.nodeCount();
  total_tables += table_used;

  if (verbose) {
    PerfMon::message(total_cycles, "total ", 
		     searcher.nodeCount());
    PerfMon::message(total_cycles, "unique", table_used);
    std::cerr << "real " << real_seconds << " sec. nps " << searcher.nodeCount()/real_seconds << "\n";
  } else {
    std::cout << filename << "\t" << searcher.nodeCount() 
	      << "\t" << table_used << "\t" << real_seconds << "\n" << std::flush;
  }
}


/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
